"""UDP server implementation for QUIC/HTTP3 connections."""

from __future__ import annotations

from typing import TYPE_CHECKING

import anyio
import anyio.abc

from .events import Event, RawData
from .task_group import TaskGroup
from .typing import AppWrapper, ConnectionState, LifespanState
from .utils import parse_socket_addr

if TYPE_CHECKING:
    from .config import Config
    from .worker_context import WorkerContext


class UDPServer:
    """Handles UDP datagrams for QUIC protocol connections."""

    def __init__(
        self,
        app: AppWrapper,
        config: Config,
        context: WorkerContext,
        state: LifespanState,
        socket: anyio.abc.UDPSocket,
    ) -> None:
        self.app = app
        self.config = config
        self.context = context
        self.socket = socket
        self.state = state

    async def run(
        self, *, task_status: anyio.abc.TaskStatus[None] = anyio.TASK_STATUS_IGNORED
    ) -> None:
        """Run the UDP server, forwarding datagrams to the QUIC protocol handler."""
        from .protocol.quic import (  # noqa: PLC0415
            QuicProtocol,  # h3/Quic is an optional part of Anycorn
        )

        task_status.started()
        server = parse_socket_addr(
            self.socket.extra(anyio.abc.SocketAttribute.raw_socket).family,  # noqa: S610
            self.socket.extra(anyio.abc.SocketAttribute.raw_socket).getsockname(),  # noqa: S610
        )
        async with TaskGroup() as task_group:
            self.protocol = QuicProtocol(
                self.app,
                self.config,
                self.context,
                task_group,
                ConnectionState(self.state.copy()),
                server,
                self.protocol_send,
            )

            while not self.context.terminated.is_set() or not self.protocol.idle:
                data, address = await self.socket.receive()
                await self.protocol.handle(RawData(data=data, address=address))

    async def protocol_send(self, event: Event) -> None:
        """Forward a protocol event back to the UDP socket."""
        if isinstance(event, RawData):
            assert event.address is not None
            await self.socket.sendto(event.data, event.address[0], event.address[1])



"""QUIC protocol handler that manages multiple HTTP/3 connections."""

from __future__ import annotations

import pathlib
from dataclasses import dataclass
from functools import partial
from typing import TYPE_CHECKING

from aioquic.buffer import Buffer
from aioquic.h3.connection import H3_ALPN
from aioquic.quic.configuration import QuicConfiguration
from aioquic.quic.connection import QuicConnection
from aioquic.quic.events import (
    ConnectionIdIssued,
    ConnectionIdRetired,
    ConnectionTerminated,
    ProtocolNegotiated,
)
from aioquic.quic.packet import (
    PACKET_TYPE_INITIAL,
    encode_quic_version_negotiation,
    pull_quic_header,
)

from anycorn.events import Closed, Event, RawData
from anycorn.typing import (
    AppWrapper,
    ConnectionState,
    SingleTask,
    TaskGroup,
    TLSExtension,
    WorkerContext,
)
from anycorn.utils import get_server_certificate_pem, tls_version_to_int

from .h3 import H3Protocol

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable

    from anycorn.config import Config

MIN_INITIAL_DATAGRAM_SIZE = 1200


@dataclass
class _Connection:
    cids: set[bytes]
    quic: QuicConnection
    task: SingleTask
    h3: H3Protocol | None = None


class QuicProtocol:
    """Manages QUIC connections and dispatches events to H3Protocol handlers."""

    def __init__(  # noqa: PLR0913
        self,
        app: AppWrapper,
        config: Config,
        context: WorkerContext,
        task_group: TaskGroup,
        state: ConnectionState,
        server: tuple[str, int] | None,
        send: Callable[[Event], Awaitable[None]],
    ) -> None:
        """Initialize the QUIC protocol handler."""
        self.app = app
        self.config = config
        self.context = context
        self.connections: dict[bytes, _Connection] = {}
        self.send = send
        self.server = server
        self.task_group = task_group
        self.state = state
        self._server_cert_pem = get_server_certificate_pem(config)

        self.quic_config = QuicConfiguration(alpn_protocols=H3_ALPN, is_client=False)
        assert config.certfile is not None
        assert config.keyfile is not None
        self.quic_config.load_cert_chain(
            certfile=pathlib.Path(config.certfile), keyfile=pathlib.Path(config.keyfile)
        )

    @property
    def idle(self) -> bool:
        """Return True when there are no active QUIC connections."""
        return len(self.connections) == 0

    async def handle(self, event: Event) -> None:
        """Handle an incoming connection event."""
        if isinstance(event, RawData):
            try:
                header = pull_quic_header(Buffer(data=event.data), host_cid_length=8)
            except ValueError:
                return
            if (
                header.version is not None
                and header.version not in self.quic_config.supported_versions
            ):
                data = encode_quic_version_negotiation(
                    source_cid=header.destination_cid,
                    destination_cid=header.source_cid,
                    supported_versions=self.quic_config.supported_versions,
                )
                await self.send(RawData(data=data, address=event.address))
                return

            connection = self.connections.get(header.destination_cid)
            if (
                connection is None
                and len(event.data) >= MIN_INITIAL_DATAGRAM_SIZE
                and header.packet_type == PACKET_TYPE_INITIAL
                and not self.context.terminated.is_set()
            ):
                quic_connection = QuicConnection(
                    configuration=self.quic_config,
                    original_destination_connection_id=header.destination_cid,
                )
                connection = _Connection(
                    cids={header.destination_cid, quic_connection.host_cid},
                    quic=quic_connection,
                    task=self.context.single_task_class(),
                )
                self.connections[header.destination_cid] = connection
                self.connections[quic_connection.host_cid] = connection

            if connection is not None:
                connection.quic.receive_datagram(event.data, event.address, now=self.context.time())
                await self._handle_events(connection, event.address)
        elif isinstance(event, Closed):
            pass

    async def send_all(self, connection: _Connection) -> None:
        """Send all pending datagrams and reschedule the connection timer."""
        for data, address in connection.quic.datagrams_to_send(now=self.context.time()):
            await self.send(RawData(data=data, address=address))

        timer = connection.quic.get_timer()
        if timer is not None:
            await connection.task.restart(
                self.task_group, partial(self._handle_timer, timer, connection)
            )

    async def _handle_events(  # noqa: C901
        self, connection: _Connection, client: tuple[str, int] | None = None
    ) -> None:
        event = connection.quic.next_event()
        while event is not None:
            if isinstance(event, ConnectionTerminated):
                await connection.task.stop()
                for cid in connection.cids:
                    del self.connections[cid]
                connection.cids = set()
            elif isinstance(event, ProtocolNegotiated):
                tls_extension = TLSExtension()
                if self._server_cert_pem is not None:
                    tls_extension["server_cert"] = self._server_cert_pem
                tls_context = getattr(connection.quic, "tls", None)
                if tls_context is not None:
                    tls_version = tls_version_to_int(getattr(tls_context, "version", None))
                    if tls_version is not None:
                        tls_extension["tls_version"] = tls_version
                    cipher_suite = getattr(tls_context, "cipher_suite", None)
                    if isinstance(cipher_suite, int):
                        tls_extension["cipher_suite"] = cipher_suite
                connection.h3 = H3Protocol(
                    self.app,
                    self.config,
                    self.context,
                    self.task_group,
                    self.state,
                    tls_extension,
                    client,
                    self.server,
                    connection.quic,
                    partial(self.send_all, connection),
                )
            elif isinstance(event, ConnectionIdIssued):
                connection.cids.add(event.connection_id)
                self.connections[event.connection_id] = connection
            elif isinstance(event, ConnectionIdRetired):
                connection.cids.remove(event.connection_id)
                del self.connections[event.connection_id]

            if connection.h3 is not None:
                await connection.h3.handle(event)

            event = connection.quic.next_event()

        await self.send_all(connection)

    async def _handle_timer(self, timer: float, connection: _Connection) -> None:
        wait = max(0, timer - self.context.time())
        await self.context.sleep(wait)
        connection.quic.handle_timer(now=self.context.time())
        await self._handle_events(connection, None)



"""Worker context and event/task wrappers for anyio-based workers."""

from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING, ClassVar

import anyio
import anyio.abc

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable

    from .typing import Event, SingleTask, TaskGroup


def _cancel_wrapper(func: Callable[[], Awaitable[None]]) -> Callable[[], Awaitable[None]]:
    @wraps(func)
    async def wrapper(
        task_status: anyio.abc.TaskStatus[anyio.CancelScope] = anyio.TASK_STATUS_IGNORED,
    ) -> None:
        cancel_scope = anyio.CancelScope()
        task_status.started(cancel_scope)
        with cancel_scope:
            await func()

    return wrapper


class AnyioSingleTask:
    """Manages a single restartable background task using an anyio CancelScope."""

    def __init__(self) -> None:
        self._handle: anyio.CancelScope | None = None
        self._lock = anyio.Lock()

    async def restart(self, task_group: TaskGroup, action: Callable) -> None:
        """Cancel any running task and start *action* as the new task."""
        async with self._lock:
            if self._handle is not None:
                self._handle.cancel()
            self._handle = await task_group._task_group.start(_cancel_wrapper(action))  # type: ignore[attr-defined]  # noqa: SLF001

    async def stop(self) -> None:
        """Cancel and clear the current task, if any."""
        async with self._lock:
            if self._handle is not None:
                self._handle.cancel()
            self._handle = None


class EventWrapper:
    """Async event primitive wrapping anyio.Event with a clear() operation."""

    def __init__(self) -> None:
        self._event = anyio.Event()

    async def clear(self) -> None:
        """Reset the event to the unset state."""
        self._event = anyio.Event()

    async def wait(self) -> None:
        """Wait until the event is set."""
        await self._event.wait()

    async def set(self) -> None:
        """Set the event, waking any waiters."""
        self._event.set()

    def is_set(self) -> bool:
        """Return True if the event has been set."""
        return self._event.is_set()


class WorkerContext:
    """Shared state for a single worker, tracking requests and shutdown signals."""

    event_class: ClassVar[type[Event]] = EventWrapper
    single_task_class: ClassVar[type[SingleTask]] = AnyioSingleTask

    def __init__(self, max_requests: int | None) -> None:
        self.max_requests = max_requests
        self.requests = 0
        self.terminate = self.event_class()
        self.terminated = self.event_class()

    async def mark_request(self) -> None:
        """Increment the request counter and signal termination if the limit is reached."""
        if self.max_requests is None:
            return

        self.requests += 1
        if self.requests > self.max_requests:
            await self.terminate.set()

    @staticmethod
    async def sleep(wait: float) -> None:
        """Sleep for *wait* seconds."""
        return await anyio.sleep(wait)

    @staticmethod
    def time() -> float:
        """Return the current event-loop time."""
        return anyio.current_time()



"""Dispatcher middleware for routing ASGI requests to multiple sub-applications."""

from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING

import anyio
import anyio.streams.memory

from anycorn.typing import ASGIFramework, ASGIReceiveEvent, Scope

if TYPE_CHECKING:
    from collections.abc import Callable

MAX_QUEUE_SIZE = 10


class _DispatcherMiddleware:
    def __init__(self, mounts: dict[str, ASGIFramework]) -> None:
        self.mounts = mounts

    async def __call__(self, scope: Scope, receive: Callable, send: Callable) -> None:
        if scope["type"] == "lifespan":
            await self._handle_lifespan(scope, receive, send)
        else:
            for path, app in self.mounts.items():
                if scope["path"].startswith(path):
                    scope["path"] = scope["path"][len(path) :] or "/"
                    return await app(scope, receive, send)
            await send(
                {
                    "type": "http.response.start",
                    "status": 404,
                    "headers": [(b"content-length", b"0")],
                }
            )
            await send({"type": "http.response.body"})
        return None

    async def _handle_lifespan(self, scope: Scope, receive: Callable, send: Callable) -> None:
        pass


class DispatcherMiddleware(_DispatcherMiddleware):
    """ASGI middleware that dispatches requests to different apps based on path prefixes."""

    async def _handle_lifespan(self, scope: Scope, receive: Callable, send: Callable) -> None:
        self.app_queues: dict[
            str,
            tuple[
                anyio.streams.memory.MemoryObjectSendStream,
                anyio.streams.memory.MemoryObjectReceiveStream,
            ],
        ] = {
            path: anyio.create_memory_object_stream[ASGIReceiveEvent](MAX_QUEUE_SIZE)
            for path in self.mounts
        }
        self.startup_complete = dict.fromkeys(self.mounts, False)
        self.shutdown_complete = dict.fromkeys(self.mounts, False)

        async with anyio.create_task_group() as tg:
            for path, app in self.mounts.items():
                tg.start_soon(
                    app,
                    scope,
                    self.app_queues[path][1].receive,
                    partial(self.send, path, send),
                )

            while True:
                message = await receive()
                for channels in self.app_queues.values():
                    await channels[0].send(message)
                if message["type"] == "lifespan.shutdown":
                    break

    async def send(self, path: str, send: Callable, message: dict) -> None:
        """Forward lifespan messages and track startup/shutdown completion across mounted apps."""
        if message["type"] == "lifespan.startup.complete":
            self.startup_complete[path] = True
            if all(self.startup_complete.values()):
                await send({"type": "lifespan.startup.complete"})
        elif message["type"] == "lifespan.shutdown.complete":
            self.shutdown_complete[path] = True
            if all(self.shutdown_complete.values()):
                await send({"type": "lifespan.shutdown.complete"})



"""Type definitions for the Anycorn ASGI/WSGI server."""

from __future__ import annotations

import sys
from collections.abc import Awaitable, Callable, Iterable, Mapping, Sequence
from multiprocessing.synchronize import Event as EventType
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Literal,
    NewType,
    Protocol,
    TypedDict,
)

import h11

from .config import Config, Sockets

if TYPE_CHECKING:
    from types import TracebackType

    import h2.events

if sys.version_info >= (3, 11):
    from typing import NotRequired, Self
else:
    from typing_extensions import NotRequired, Self

H11SendableEvent = h11.Data | h11.EndOfMessage | h11.InformationalResponse | h11.Response

WorkerFunc = Callable[[Config, Sockets | None, EventType | None], None]

LifespanState = dict[str, Any]

ConnectionState = NewType("ConnectionState", dict[str, Any])


class ASGIVersions(TypedDict, total=False):
    """ASGI version information."""

    spec_version: str
    version: Literal["2.0", "3.0"]


class TLSExtension(TypedDict, total=False):
    """TLS connection extension data for the ASGI connection scope."""

    server_cert: str | None
    client_cert_chain: Sequence[str]
    client_cert_name: str | None
    client_cert_error: str | None
    tls_version: int | None
    cipher_suite: int | None


Extensions = TypedDict(
    "Extensions",
    {
        "tls": TLSExtension,
        "http.response.push": Mapping[str, Any],
        "http.response.trailers": Mapping[str, Any],
        "http.response.early_hint": Mapping[str, Any],
        "websocket.http.response": Mapping[str, Any],
    },
    total=False,
)


class HTTPScope(TypedDict):
    """ASGI HTTP connection scope."""

    type: Literal["http"]
    asgi: ASGIVersions
    http_version: str
    method: str
    scheme: str
    path: str
    raw_path: bytes
    query_string: bytes
    root_path: str
    headers: Iterable[tuple[bytes, bytes]]
    client: tuple[str, int] | None
    server: tuple[str, int | None] | None
    state: ConnectionState
    extensions: Extensions


class WebsocketScope(TypedDict):
    """ASGI WebSocket connection scope."""

    type: Literal["websocket"]
    asgi: ASGIVersions
    http_version: str
    scheme: str
    path: str
    raw_path: bytes
    query_string: bytes
    root_path: str
    headers: Iterable[tuple[bytes, bytes]]
    client: tuple[str, int] | None
    server: tuple[str, int | None] | None
    subprotocols: Iterable[str]
    state: ConnectionState
    extensions: Extensions


class LifespanScope(TypedDict):
    """ASGI Lifespan connection scope."""

    type: Literal["lifespan"]
    asgi: ASGIVersions
    state: LifespanState


WWWScope = HTTPScope | WebsocketScope
Scope = HTTPScope | WebsocketScope | LifespanScope


class HTTPRequestEvent(TypedDict):
    """ASGI HTTP request receive event."""

    type: Literal["http.request"]
    body: bytes
    more_body: bool


class HTTPResponseStartEvent(TypedDict):
    """ASGI HTTP response start send event."""

    type: Literal["http.response.start"]
    status: int
    headers: Iterable[tuple[bytes, bytes]]
    trailers: NotRequired[bool]


class HTTPResponseBodyEvent(TypedDict):
    """ASGI HTTP response body send event."""

    type: Literal["http.response.body"]
    body: bytes
    more_body: bool


class HTTPResponseTrailersEvent(TypedDict):
    """ASGI HTTP response trailers send event."""

    type: Literal["http.response.trailers"]
    headers: Iterable[tuple[bytes, bytes]]
    more_trailers: NotRequired[bool]


class HTTPServerPushEvent(TypedDict):
    """ASGI HTTP server push send event."""

    type: Literal["http.response.push"]
    path: str
    headers: Iterable[tuple[bytes, bytes]]


class HTTPEarlyHintEvent(TypedDict):
    """ASGI HTTP early hint send event."""

    type: Literal["http.response.early_hint"]
    links: Iterable[bytes]


class HTTPDisconnectEvent(TypedDict):
    """ASGI HTTP disconnect receive event."""

    type: Literal["http.disconnect"]


class WebsocketConnectEvent(TypedDict):
    """ASGI WebSocket connect receive event."""

    type: Literal["websocket.connect"]


class WebsocketAcceptEvent(TypedDict):
    """ASGI WebSocket accept send event."""

    type: Literal["websocket.accept"]
    subprotocol: str | None
    headers: Iterable[tuple[bytes, bytes]]


class WebsocketReceiveEvent(TypedDict):
    """ASGI WebSocket receive event."""

    type: Literal["websocket.receive"]
    bytes: bytes | None
    text: str | None


class WebsocketSendEvent(TypedDict):
    """ASGI WebSocket send event."""

    type: Literal["websocket.send"]
    bytes: bytes | None
    text: str | None


class WebsocketResponseStartEvent(TypedDict):
    """ASGI WebSocket HTTP response start event."""

    type: Literal["websocket.http.response.start"]
    status: int
    headers: Iterable[tuple[bytes, bytes]]


class WebsocketResponseBodyEvent(TypedDict):
    """ASGI WebSocket HTTP response body event."""

    type: Literal["websocket.http.response.body"]
    body: bytes
    more_body: bool


class WebsocketDisconnectEvent(TypedDict):
    """ASGI WebSocket disconnect receive event."""

    type: Literal["websocket.disconnect"]
    code: int


class WebsocketCloseEvent(TypedDict):
    """ASGI WebSocket close send event."""

    type: Literal["websocket.close"]
    code: int
    reason: str | None


class LifespanStartupEvent(TypedDict):
    """ASGI Lifespan startup receive event."""

    type: Literal["lifespan.startup"]


class LifespanShutdownEvent(TypedDict):
    """ASGI Lifespan shutdown receive event."""

    type: Literal["lifespan.shutdown"]


class LifespanStartupCompleteEvent(TypedDict):
    """ASGI Lifespan startup complete send event."""

    type: Literal["lifespan.startup.complete"]


class LifespanStartupFailedEvent(TypedDict):
    """ASGI Lifespan startup failed send event."""

    type: Literal["lifespan.startup.failed"]
    message: str


class LifespanShutdownCompleteEvent(TypedDict):
    """ASGI Lifespan shutdown complete send event."""

    type: Literal["lifespan.shutdown.complete"]


class LifespanShutdownFailedEvent(TypedDict):
    """ASGI Lifespan shutdown failed send event."""

    type: Literal["lifespan.shutdown.failed"]
    message: str


ASGIReceiveEvent = (
    HTTPRequestEvent
    | HTTPDisconnectEvent
    | WebsocketConnectEvent
    | WebsocketReceiveEvent
    | WebsocketDisconnectEvent
    | LifespanStartupEvent
    | LifespanShutdownEvent
)


ASGISendEvent = (
    HTTPResponseStartEvent
    | HTTPResponseBodyEvent
    | HTTPResponseTrailersEvent
    | HTTPServerPushEvent
    | HTTPEarlyHintEvent
    | HTTPDisconnectEvent
    | WebsocketAcceptEvent
    | WebsocketSendEvent
    | WebsocketResponseStartEvent
    | WebsocketResponseBodyEvent
    | WebsocketCloseEvent
    | LifespanStartupCompleteEvent
    | LifespanStartupFailedEvent
    | LifespanShutdownCompleteEvent
    | LifespanShutdownFailedEvent
)


ASGIReceiveCallable = Callable[[], Awaitable[ASGIReceiveEvent]]
ASGISendCallable = Callable[[ASGISendEvent], Awaitable[None]]

ASGIFramework = Callable[
    [
        Scope,
        ASGIReceiveCallable,
        ASGISendCallable,
    ],
    Awaitable[None],
]
WSGIFramework = Callable[[dict, Callable], Iterable[bytes]]
Framework = ASGIFramework | WSGIFramework


class H2SyncStream(Protocol):
    """Protocol for synchronous HTTP/2 stream handling."""

    scope: dict

    def data_received(self, data: bytes) -> None:
        """Handle received data for this stream."""
        ...

    def ended(self) -> None:
        """Signal that the stream has ended."""
        ...

    def reset(self) -> None:
        """Reset the stream."""
        ...

    def close(self) -> None:
        """Close the stream."""
        ...

    async def handle_request(
        self,
        event: h2.events.RequestReceived,
        scheme: str,
        client: tuple[str, int],
        server: tuple[str, int],
    ) -> None:
        """Handle an incoming HTTP/2 request event."""
        ...


class H2AsyncStream(Protocol):
    """Protocol for asynchronous HTTP/2 stream handling."""

    scope: dict

    async def data_received(self, data: bytes) -> None:
        """Handle received data for this stream."""
        ...

    async def ended(self) -> None:
        """Signal that the stream has ended."""
        ...

    async def reset(self) -> None:
        """Reset the stream."""
        ...

    async def close(self) -> None:
        """Close the stream."""
        ...

    async def handle_request(
        self,
        event: h2.events.RequestReceived,
        scheme: str,
        client: tuple[str, int],
        server: tuple[str, int],
    ) -> None:
        """Handle an incoming HTTP/2 request event."""
        ...


class Event(Protocol):
    """Protocol for async event primitives."""

    def __init__(self) -> None: ...

    async def clear(self) -> None:
        """Clear the event, resetting it to the unset state."""
        ...

    async def set(self) -> None:
        """Set the event, unblocking any waiters."""
        ...

    async def wait(self) -> None:
        """Wait until the event is set."""
        ...

    def is_set(self) -> bool:
        """Return True if the event is currently set."""
        ...


class WorkerContext(Protocol):
    """Protocol describing the shared state passed to each connection handler."""

    event_class: ClassVar[type[Event]]
    single_task_class: ClassVar[type[SingleTask]]
    terminate: Event
    terminated: Event

    async def mark_request(self) -> None:
        """Record that a request has been received."""
        ...

    @staticmethod
    async def sleep(wait: float) -> None:
        """Sleep for the given number of seconds."""
        ...

    @staticmethod
    def time() -> float:
        """Return the current time as a float."""
        ...


class TaskGroup(Protocol):
    """Protocol for the task group used to manage concurrent ASGI tasks."""

    async def spawn_app(
        self,
        app: AppWrapper,
        config: Config,
        scope: Scope,
        send: Callable[[ASGISendEvent | None], Awaitable[None]],
    ) -> Callable[[ASGIReceiveEvent], Awaitable[None]]:
        """Spawn the ASGI app as a task and return a callable to send receive events."""
        ...

    def spawn(self, func: Callable, *args: Any) -> None:  # noqa: ANN401
        """Spawn a background task running func with the given arguments."""
        ...

    async def __aenter__(self) -> Self: ...

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        tb: TracebackType | None,
    ) -> None: ...


class ResponseSummary(TypedDict):
    """Minimal HTTP response summary used for access logging."""

    status: int
    headers: Iterable[tuple[bytes, bytes]]


class AppWrapper(Protocol):
    """Protocol for application wrappers (ASGI or WSGI)."""

    async def __call__(
        self,
        scope: Scope,
        receive: ASGIReceiveCallable,
        send: ASGISendCallable,
        sync_spawn: Callable,
        call_soon: Callable,
    ) -> None:
        """Invoke the wrapped application."""
        ...


class SingleTask(Protocol):
    """Protocol for managing a single restartable background task."""

    def __init__(self) -> None: ...

    async def restart(self, task_group: TaskGroup, action: Callable) -> None:
        """Restart the background task using the given task group and action callable."""
        ...

    async def stop(self) -> None:
        """Stop the background task."""
        ...



"""Entry points for running Anycorn workers."""

from __future__ import annotations

import platform
import signal
import sys
import time
from functools import partial
from multiprocessing import get_context
from multiprocessing.connection import wait
from pickle import PicklingError
from random import randint
from typing import TYPE_CHECKING, Any

import anyio
import anyio.abc
import anyio.streams.tls

from .lifespan import Lifespan
from .statsd import StatsdLogger
from .tcp_server import tcp_server_handler
from .typing import AppWrapper, ConnectionState, LifespanState, WorkerFunc
from .utils import (
    ShutdownError,
    check_for_updates,
    check_multiprocess_shutdown_event,
    files_to_watch,
    load_application,
    raise_shutdown,
    repr_socket_addr,
    write_pid_file,
)
from .worker_context import WorkerContext

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable
    from multiprocessing.context import BaseContext
    from multiprocessing.process import BaseProcess
    from multiprocessing.synchronize import Event as EventType

    from .config import Config, Sockets

if sys.version_info < (3, 11):
    from exceptiongroup import BaseExceptionGroup


def run(config: Config) -> int:  # noqa: C901, PLR0912
    """Start the server, blocking until it exits, and return an exit code."""
    if config.pid_path is not None:
        write_pid_file(config.pid_path)

    worker_func: WorkerFunc
    worker_func = anyio_worker

    sockets = config.create_sockets()

    if config.use_reloader and config.workers == 0:
        msg = "Cannot reload without workers"
        raise RuntimeError(msg)

    exitcode = 0
    if config.workers == 0:
        worker_func(config, sockets)
    else:
        if config.use_reloader:
            # Load the application so that the correct paths are checked for
            # changes, but only when the reloader is being used.
            load_application(config.application_path, config.wsgi_max_body_size)

        ctx = get_context("spawn")

        active = True
        shutdown_event = ctx.Event()

        def shutdown(*_args: Any) -> None:  # noqa: ANN401
            nonlocal active, shutdown_event
            shutdown_event.set()
            active = False

        processes: list[BaseProcess] = []
        while active:
            # Ignore SIGINT before creating the processes, so that they
            # inherit the signal handling. This means that the shutdown
            # function controls the shutdown.
            signal.signal(signal.SIGINT, signal.SIG_IGN)

            _populate(processes, config, worker_func, sockets, shutdown_event, ctx)

            for signal_name in ("SIGINT", "SIGTERM", "SIGBREAK"):
                if hasattr(signal, signal_name):
                    signal.signal(getattr(signal, signal_name), shutdown)

            if config.use_reloader:
                files = files_to_watch()
                while True:
                    finished = wait((process.sentinel for process in processes), timeout=1)
                    updated = check_for_updates(files)
                    if updated:
                        shutdown_event.set()
                        for process in processes:
                            process.join()
                        shutdown_event.clear()
                        break
                    if len(finished) > 0:
                        break
            else:
                wait(process.sentinel for process in processes)

            exitcode = _join_exited(processes)
            if exitcode != 0:
                shutdown_event.set()
                active = False

        for process in processes:
            process.terminate()

        exitcode = _join_exited(processes) if exitcode != 0 else exitcode

        for sock in sockets.secure_sockets:
            sock.close()

        for sock in sockets.insecure_sockets:
            sock.close()

    return exitcode


def _populate(  # noqa: PLR0913
    processes: list[BaseProcess],
    config: Config,
    worker_func: WorkerFunc,
    sockets: Sockets,
    shutdown_event: EventType,
    ctx: BaseContext,
) -> None:
    for _ in range(config.workers - len(processes)):
        process = ctx.Process(  # type: ignore[attr-defined]
            target=worker_func,
            kwargs={"config": config, "shutdown_event": shutdown_event, "sockets": sockets},
        )
        process.daemon = True
        try:
            process.start()
        except PicklingError as error:
            msg = "Cannot pickle the config, see https://docs.python.org/3/library/pickle.html#pickle-picklable"
            raise RuntimeError(msg) from error
        processes.append(process)
        if platform.system() == "Windows":
            time.sleep(0.1)


def _join_exited(processes: list[BaseProcess]) -> int:
    exitcode = 0
    for index in reversed(range(len(processes))):
        worker = processes[index]
        if worker.exitcode is not None:
            worker.join()
            exitcode = worker.exitcode if exitcode == 0 else exitcode
            del processes[index]

    return exitcode


async def worker_serve(  # noqa: C901, PLR0915
    app: AppWrapper,
    config: Config,
    *,
    sockets: Sockets | None = None,
    shutdown_trigger: Callable[..., Awaitable[None]] | None = None,
    task_status: anyio.abc.TaskStatus[list[str]] = anyio.TASK_STATUS_IGNORED,
) -> None:
    """Run the server workers, handling lifespan and connections."""
    config.set_statsd_logger_class(StatsdLogger)

    lifespan_state: LifespanState = {}
    lifespan = Lifespan(app, config, lifespan_state)
    max_requests = None
    if config.max_requests is not None:
        max_requests = config.max_requests + randint(0, config.max_requests_jitter)  # noqa: S311
    context = WorkerContext(max_requests)

    async with anyio.create_task_group() as lifespan_tg:
        await lifespan_tg.start(lifespan.handle_lifespan)
        await lifespan.wait_for_startup()

        async with anyio.create_task_group() as server_tg:
            if sockets is None:
                sockets = config.create_sockets()
                for sock in sockets.secure_sockets:
                    sock.listen(config.backlog)
                for sock in sockets.insecure_sockets:
                    sock.listen(config.backlog)

            ssl_context = config.create_ssl_context()
            listeners: list[anyio.abc.SocketListener | anyio.streams.tls.TLSListener] = []
            binds = []
            for secure_sock in sockets.secure_sockets:
                assert ssl_context is not None
                asynclib = anyio._core._eventloop.get_async_backend()  # noqa: SLF001  # ty:ignore[possibly-missing-attribute]
                secure_listener = anyio.streams.tls.TLSListener(
                    asynclib.create_tcp_listener(secure_sock),
                    ssl_context,
                    True,  # noqa: FBT003
                    config.ssl_handshake_timeout,
                )
                listeners.append(secure_listener)
                bind = repr_socket_addr(secure_sock.family, secure_sock.getsockname())
                url = f"https://{bind}"
                binds.append(url)
                await config.log.info("Running on %s (CTRL + C to quit)", url)

            for insecure_sock in sockets.insecure_sockets:
                asynclib = anyio._core._eventloop.get_async_backend()  # noqa: SLF001  # ty:ignore[possibly-missing-attribute]
                insecure_listener = asynclib.create_tcp_listener(insecure_sock)
                listeners.append(insecure_listener)
                bind = repr_socket_addr(insecure_sock.family, insecure_sock.getsockname())
                url = f"http://{bind}"
                binds.append(url)
                await config.log.info("Running on %s (CTRL + C to quit)", url)

            task_status.started(binds)
            try:
                async with anyio.create_task_group() as tg:
                    if shutdown_trigger is not None:
                        tg.start_soon(raise_shutdown, shutdown_trigger)
                    tg.start_soon(raise_shutdown, context.terminate.wait)

                    for listener in listeners:
                        tg.start_soon(
                            partial(
                                listener.serve,
                                tcp_server_handler(
                                    app, config, context, ConnectionState(lifespan_state.copy())
                                ),
                            ),
                        )

                    await anyio.Event().wait()
            except BaseExceptionGroup as error:
                _, other_errors = error.split((ShutdownError, KeyboardInterrupt))
                if other_errors is not None:
                    raise other_errors from error
            finally:
                await context.terminated.set()
                server_tg.cancel_scope.deadline = anyio.current_time() + config.graceful_timeout

        await lifespan.wait_for_shutdown()
        lifespan_tg.cancel_scope.cancel()


def anyio_worker(
    config: Config, sockets: Sockets | None = None, shutdown_event: EventType | None = None
) -> None:
    """Run the anyio worker, loading the application and serving requests."""
    if sockets is not None:
        for sock in sockets.secure_sockets:
            sock.listen(config.backlog)
        for sock in sockets.insecure_sockets:
            sock.listen(config.backlog)
    app = load_application(config.application_path, config.wsgi_max_body_size)

    shutdown_trigger = None
    if shutdown_event is not None:
        shutdown_trigger = partial(check_multiprocess_shutdown_event, shutdown_event, anyio.sleep)

    anyio.run(
        partial(worker_serve, app, config, sockets=sockets, shutdown_trigger=shutdown_trigger),
        backend=config.worker_class,
    )



"""ASGI WebSocket stream handler that manages handshake and message framing."""

from __future__ import annotations

from enum import Enum, auto
from io import BytesIO, StringIO
from time import time
from typing import TYPE_CHECKING
from urllib.parse import unquote

from wsproto.connection import Connection, ConnectionState, ConnectionType
from wsproto.events import (
    BytesMessage,
    CloseConnection,
    Message,
    Ping,
    TextMessage,
)
from wsproto.events import (
    Event as WSProtoEvent,
)
from wsproto.extensions import Extension, PerMessageDeflate
from wsproto.frame_protocol import CloseReason
from wsproto.handshake import WEBSOCKET_VERSION, server_extensions_handshake
from wsproto.utilities import LocalProtocolError, generate_accept_token, split_comma_header

from anycorn.typing import (
    AppWrapper,
    ASGISendEvent,
    Extensions,
    TaskGroup,
    TLSExtension,
    WebsocketAcceptEvent,
    WebsocketResponseBodyEvent,
    WebsocketResponseStartEvent,
    WebsocketScope,
    WorkerContext,
)
from anycorn.utils import (
    UnexpectedMessageError,
    build_and_validate_headers,
    suppress_body,
    valid_server_name,
)

from .events import Body, Data, EndBody, EndData, Event, Request, Response, StreamClosed

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable, Iterable

    from anycorn.config import Config


class ASGIWebsocketState(Enum):
    """State machine for ASGI WebSocket lifecycle.

    Hypercorn supports the ASGI websocket HTTP response extension,
    which allows HTTP responses rather than acceptance.
    """

    HANDSHAKE = auto()
    CONNECTED = auto()
    RESPONSE = auto()
    CLOSED = auto()
    HTTPCLOSED = auto()


class FrameTooLargeError(Exception):
    """Raised when a WebSocket message exceeds the configured maximum size."""


class Handshake:
    """Parses and validates a WebSocket upgrade handshake."""

    def __init__(
        self,
        headers: list[tuple[bytes, bytes]],
        http_version: str,
        websocket_permessage_deflate: bool = False,  # noqa: FBT001,FBT002
    ) -> None:
        """Parse WebSocket-related headers from the request."""
        self.accepted = False
        self.http_version = http_version
        self.websocket_permessage_deflate = websocket_permessage_deflate
        self.connection_tokens: list[str] | None = None
        self.extensions: list[str] | None = None
        self.key: bytes | None = None
        self.subprotocols: list[str] | None = None
        self.upgrade: bytes | None = None
        self.version: bytes | None = None
        for raw_name, value in headers:
            name = raw_name.lower()
            if name == b"connection":
                self.connection_tokens = split_comma_header(value)
            elif name == b"sec-websocket-extensions":
                self.extensions = split_comma_header(value)
            elif name == b"sec-websocket-key":
                self.key = value
            elif name == b"sec-websocket-protocol":
                self.subprotocols = split_comma_header(value)
            elif name == b"sec-websocket-version":
                self.version = value
            elif name == b"upgrade":
                self.upgrade = value

    def is_valid(self) -> bool:
        """Return True if the handshake headers form a valid WebSocket upgrade."""
        if self.http_version < "1.1":
            return False
        if self.http_version == "1.1":
            if self.key is None:
                return False
            if self.connection_tokens is None or not any(
                token.lower() == "upgrade" for token in self.connection_tokens
            ):
                return False
            if self.upgrade is None or self.upgrade.lower() != b"websocket":
                return False

        return self.version == WEBSOCKET_VERSION

    def accept(
        self,
        subprotocol: str | None,
        additional_headers: Iterable[tuple[bytes, bytes]],
    ) -> tuple[int, list[tuple[bytes, bytes]], Connection]:
        """Build accept headers and return the status code, headers, and connection."""
        headers = []
        if subprotocol is not None:
            if self.subprotocols is None or subprotocol not in self.subprotocols:
                msg = "Invalid Subprotocol"
                raise Exception(msg)  # noqa: TRY002
            headers.append((b"sec-websocket-protocol", subprotocol.encode()))

        extensions: list[Extension] = []
        if self.websocket_permessage_deflate:
            extensions.append(PerMessageDeflate())
        accepts = None
        if self.extensions is not None:
            accepts = server_extensions_handshake(self.extensions, extensions)

        if accepts:
            headers.append((b"sec-websocket-extensions", accepts))

        if self.key is not None:
            headers.append((b"sec-websocket-accept", generate_accept_token(self.key)))

        status_code = 200
        if self.http_version == "1.1":
            headers.extend([(b"upgrade", b"WebSocket"), (b"connection", b"Upgrade")])
            status_code = 101

        for name, value in additional_headers:
            if name == b"sec-websocket-protocol" or name.startswith(b":"):
                msg = f"Invalid additional header, {name.decode()}"
                raise Exception(msg)  # noqa: TRY002

            headers.append((name, value))

        self.accepted = True
        return status_code, headers, Connection(ConnectionType.SERVER, extensions)


class WebsocketBuffer:
    """Accumulates WebSocket message fragments into a single message."""

    def __init__(self, max_length: int) -> None:
        """Initialize with the maximum allowed message length in bytes."""
        self.value: BytesIO | StringIO | None = None
        self.length = 0
        self.max_length = max_length

    def extend(self, event: Message) -> None:
        """Append a message fragment, raising FrameTooLargeError if the limit is exceeded."""
        if self.value is None:
            if isinstance(event, TextMessage):
                self.value = StringIO()
            else:
                self.value = BytesIO()
        self.length += self.value.write(event.data)
        if self.length > self.max_length:
            raise FrameTooLargeError

    def clear(self) -> None:
        """Reset the buffer ready for the next message."""
        self.value = None
        self.length = 0

    def to_message(self) -> dict:
        """Return the accumulated data as a websocket.receive ASGI message."""
        return {
            "type": "websocket.receive",
            "bytes": self.value.getvalue() if isinstance(self.value, BytesIO) else None,
            "text": self.value.getvalue() if isinstance(self.value, StringIO) else None,
        }


class WSStream:
    """Handles a single WebSocket stream including handshake and message framing."""

    def __init__(  # noqa: PLR0913
        self,
        app: AppWrapper,
        config: Config,
        context: WorkerContext,
        task_group: TaskGroup,
        client: tuple[str, int] | None,
        server: tuple[str, int] | None,
        send: Callable[[Event], Awaitable[None]],
        stream_id: int,
        tls: TLSExtension | None,
    ) -> None:
        """Initialize the WebSocket stream handler."""
        self.app = app
        self.app_put: Callable | None = None
        self.buffer = WebsocketBuffer(config.websocket_max_message_size)
        self.client = client
        self.closed = False
        self.config = config
        self.context = context
        self.task_group = task_group
        self.response: WebsocketResponseStartEvent
        self.scope: WebsocketScope
        self.send = send
        self.server = server
        self.start_time: float
        self.state = ASGIWebsocketState.HANDSHAKE
        self.stream_id = stream_id
        self.tls = tls
        self.scheme = "wss" if self.tls is not None else "ws"

        self.connection: Connection
        self.handshake: Handshake

    @property
    def idle(self) -> bool:
        """Return True when the WebSocket stream is in a terminal state."""
        return self.state in {ASGIWebsocketState.CLOSED, ASGIWebsocketState.HTTPCLOSED}

    async def handle(self, event: Event) -> None:  # noqa: C901
        """Handle an incoming stream event."""
        if self.closed:
            return
        if isinstance(event, Request):
            self.start_time = time()
            self.handshake = Handshake(
                event.headers, event.http_version, self.config.websocket_permessage_deflate
            )
            path, _, query_string = event.raw_path.partition(b"?")
            extensions = Extensions()
            if self.tls is not None:
                extensions["tls"] = self.tls
            extensions["websocket.http.response"] = {}
            self.scope = {
                "type": "websocket",
                "asgi": {"spec_version": "2.3", "version": "3.0"},
                "scheme": self.scheme,
                "http_version": event.http_version,
                "path": unquote(path.decode("ascii")),
                "raw_path": path,
                "query_string": query_string,
                "root_path": self.config.root_path,
                "headers": event.headers,
                "client": self.client,
                "server": self.server,
                "state": event.state,
                "subprotocols": self.handshake.subprotocols or [],
                "extensions": extensions,
            }

            if not valid_server_name(self.config, event):
                await self._send_error_response(404)
                self.closed = True
            elif not self.handshake.is_valid():
                await self._send_error_response(400)
                self.closed = True
            else:
                self.app_put = await self.task_group.spawn_app(
                    self.app, self.config, self.scope, self.app_send
                )
                await self.app_put({"type": "websocket.connect"})
        elif isinstance(event, (Body, Data)) and not self.handshake.accepted:
            await self._send_error_response(400)
            self.closed = True
        elif isinstance(event, (Body, Data)):
            self.connection.receive_data(event.data)
            await self._handle_events()
        elif isinstance(event, StreamClosed):
            self.closed = True
            if self.app_put is not None:
                if self.state in {ASGIWebsocketState.HTTPCLOSED, ASGIWebsocketState.CLOSED}:
                    code = CloseReason.NORMAL_CLOSURE.value
                else:
                    code = CloseReason.ABNORMAL_CLOSURE.value
                await self.app_put({"type": "websocket.disconnect", "code": code})

    async def app_send(self, message: ASGISendEvent | None) -> None:  # noqa: C901, PLR0912
        """Handle a message sent by the ASGI application."""
        if self.closed:
            # Allow app to finish after close
            return

        if message is None:  # ASGI App has finished sending messages
            # Cleanup if required
            if self.state == ASGIWebsocketState.HANDSHAKE:
                await self._send_error_response(500)
                await self.config.log.access(
                    self.scope, {"status": 500, "headers": []}, time() - self.start_time
                )
            elif self.state == ASGIWebsocketState.CONNECTED:
                await self._send_wsproto_event(CloseConnection(code=CloseReason.INTERNAL_ERROR))
            await self.send(StreamClosed(stream_id=self.stream_id))
        elif message["type"] == "websocket.accept" and self.state == ASGIWebsocketState.HANDSHAKE:
            await self._accept(message)
        elif (
            message["type"] == "websocket.http.response.start"
            and self.state == ASGIWebsocketState.HANDSHAKE
        ):
            self.response = message
        elif message["type"] == "websocket.http.response.body" and self.state in {
            ASGIWebsocketState.HANDSHAKE,
            ASGIWebsocketState.RESPONSE,
        }:
            await self._send_rejection(message)
        elif message["type"] == "websocket.send" and self.state == ASGIWebsocketState.CONNECTED:
            event: WSProtoEvent
            msg_bytes = message.get("bytes")
            if msg_bytes is not None:
                event = BytesMessage(data=bytes(msg_bytes))
            elif not isinstance(message["text"], str):
                msg = f"{message['text']} should be a str"
                raise TypeError(msg)
            else:
                event = TextMessage(data=message["text"])
            await self._send_wsproto_event(event)
        elif message["type"] == "websocket.close" and self.state == ASGIWebsocketState.HANDSHAKE:
            self.state = ASGIWebsocketState.HTTPCLOSED
            await self._send_error_response(403)
        elif message["type"] == "websocket.close":
            self.state = ASGIWebsocketState.CLOSED
            await self._send_wsproto_event(
                CloseConnection(
                    code=int(message.get("code", CloseReason.NORMAL_CLOSURE)),
                    reason=message.get("reason"),
                )
            )
            await self.send(EndData(stream_id=self.stream_id))
        else:
            raise UnexpectedMessageError(self.state, message["type"])

    async def _handle_events(self) -> None:
        for event in self.connection.events():
            if isinstance(event, Message):
                try:
                    self.buffer.extend(event)
                except FrameTooLargeError:
                    await self._send_wsproto_event(
                        CloseConnection(code=CloseReason.MESSAGE_TOO_BIG)
                    )
                    break

                if event.message_finished:
                    assert self.app_put is not None
                    await self.app_put(self.buffer.to_message())
                    self.buffer.clear()
            elif isinstance(event, Ping):
                await self._send_wsproto_event(event.response())
            elif isinstance(event, CloseConnection):
                if self.connection.state == ConnectionState.REMOTE_CLOSING:
                    await self._send_wsproto_event(event.response())
                await self.send(StreamClosed(stream_id=self.stream_id))

    async def _send_error_response(self, status_code: int) -> None:
        await self.send(
            Response(
                stream_id=self.stream_id,
                status_code=status_code,
                headers=[(b"content-length", b"0"), (b"connection", b"close")],
            )
        )
        await self.send(EndBody(stream_id=self.stream_id))
        await self.config.log.access(
            self.scope, {"status": status_code, "headers": []}, time() - self.start_time
        )

    async def _send_wsproto_event(self, event: WSProtoEvent) -> None:
        try:
            data = self.connection.send(event)
        except LocalProtocolError:
            pass
        else:
            await self.send(Data(stream_id=self.stream_id, data=data))

    async def _accept(self, message: WebsocketAcceptEvent) -> None:
        self.state = ASGIWebsocketState.CONNECTED
        status_code, headers, self.connection = self.handshake.accept(
            message.get("subprotocol"), message.get("headers", [])
        )
        await self.send(
            Response(stream_id=self.stream_id, status_code=status_code, headers=headers)
        )
        await self.config.log.access(
            self.scope, {"status": status_code, "headers": []}, time() - self.start_time
        )
        if self.config.websocket_ping_interval is not None:
            self.task_group.spawn(self._send_pings)

    async def _send_rejection(self, message: WebsocketResponseBodyEvent) -> None:
        body_suppressed = suppress_body("GET", self.response["status"])
        if self.state == ASGIWebsocketState.HANDSHAKE:
            headers = build_and_validate_headers(self.response["headers"])
            await self.send(
                Response(
                    stream_id=self.stream_id,
                    status_code=int(self.response["status"]),
                    headers=headers,
                )
            )
            self.state = ASGIWebsocketState.RESPONSE
        if not body_suppressed:
            await self.send(Body(stream_id=self.stream_id, data=bytes(message.get("body", b""))))
        if not message.get("more_body", False):
            self.state = ASGIWebsocketState.HTTPCLOSED
            await self.send(EndBody(stream_id=self.stream_id))
            await self.config.log.access(self.scope, self.response, time() - self.start_time)

    async def _send_pings(self) -> None:
        ping_interval = self.config.websocket_ping_interval
        assert ping_interval is not None
        while not self.closed:
            await self._send_wsproto_event(Ping())
            await self.context.sleep(ping_interval)



"""Internal event types used for communication between server components."""

from __future__ import annotations

from abc import ABC
from dataclasses import dataclass


class Event(ABC):  # noqa: B024
    """Base class for all internal server events."""


@dataclass(frozen=True)
class RawData(Event):
    """Event carrying raw bytes received from the network."""

    data: bytes
    address: tuple[str, int] | None = None


@dataclass(frozen=True)
class Closed(Event):
    """Event signalling that a connection has been closed."""


@dataclass(frozen=True)
class Updated(Event):
    """Event signalling a protocol state change."""

    idle: bool



"""HTTP/3 protocol implementation built on aioquic."""

from __future__ import annotations

from typing import TYPE_CHECKING

from aioquic.h3.connection import H3Connection
from aioquic.h3.events import DataReceived, HeadersReceived
from aioquic.h3.exceptions import NoAvailablePushIDError

from anycorn.utils import filter_pseudo_headers

from .events import (
    Body,
    Data,
    EndBody,
    EndData,
    InformationalResponse,
    Request,
    Response,
    StreamClosed,
    Trailers,
)
from .events import (
    Event as StreamEvent,
)
from .http_stream import HTTPStream
from .ws_stream import WSStream

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable

    from aioquic.quic.connection import QuicConnection
    from aioquic.quic.events import QuicEvent

    from anycorn.config import Config
    from anycorn.typing import AppWrapper, ConnectionState, TaskGroup, TLSExtension, WorkerContext


class H3Protocol:
    """HTTP/3 server-side protocol handler using aioquic."""

    def __init__(  # noqa: PLR0913
        self,
        app: AppWrapper,
        config: Config,
        context: WorkerContext,
        task_group: TaskGroup,
        state: ConnectionState,
        tls: TLSExtension,
        client: tuple[str, int] | None,
        server: tuple[str, int] | None,
        quic: QuicConnection,
        send: Callable[[], Awaitable[None]],
    ) -> None:
        """Initialize the H3 protocol handler."""
        self.app = app
        self.client = client
        self.config = config
        self.context = context
        self.connection = H3Connection(quic)
        self.send = send
        self.server = server
        self.streams: dict[int, HTTPStream | WSStream] = {}
        self.task_group = task_group
        self.state = state
        self.tls = tls

    async def handle(self, quic_event: QuicEvent) -> None:
        """Handle an incoming QUIC event."""
        for event in self.connection.handle_event(quic_event):
            if isinstance(event, HeadersReceived):
                if not self.context.terminated.is_set():
                    await self._create_stream(event)
                    if event.stream_ended:
                        await self.streams[event.stream_id].handle(
                            EndBody(stream_id=event.stream_id)
                        )
            elif isinstance(event, DataReceived):
                await self.streams[event.stream_id].handle(
                    Body(stream_id=event.stream_id, data=event.data)
                )
                if event.stream_ended:
                    await self.streams[event.stream_id].handle(EndBody(stream_id=event.stream_id))

    async def stream_send(self, event: StreamEvent) -> None:
        """Send a stream event to the remote client."""
        if isinstance(event, (InformationalResponse, Response)):
            self.connection.send_headers(
                event.stream_id,
                [
                    (b":status", b"%d" % event.status_code),
                    *event.headers,
                    *self.config.response_headers("h3"),
                ],
            )
            await self.send()
        elif isinstance(event, (Body, Data)):
            self.connection.send_data(event.stream_id, event.data, False)  # noqa: FBT003
            await self.send()
        elif isinstance(event, (EndBody, EndData)):
            self.connection.send_data(event.stream_id, b"", True)  # noqa: FBT003
            await self.send()
        elif isinstance(event, Trailers):
            self.connection.send_headers(event.stream_id, event.headers)
            await self.send()
        elif isinstance(event, StreamClosed):
            pass  # ??
        elif isinstance(event, Request):
            await self._create_server_push(event.stream_id, event.raw_path, event.headers)

    async def _create_stream(self, request: HeadersReceived) -> None:
        for name, value in request.headers:
            if name == b":method":
                method = value.decode("ascii").upper()
            elif name == b":path":
                raw_path = value

        if method == "CONNECT":
            self.streams[request.stream_id] = WSStream(
                self.app,
                self.config,
                self.context,
                self.task_group,
                self.client,
                self.server,
                self.stream_send,
                request.stream_id,
                self.tls,
            )
        else:
            self.streams[request.stream_id] = HTTPStream(
                self.app,
                self.config,
                self.context,
                self.task_group,
                self.client,
                self.server,
                self.stream_send,
                request.stream_id,
                self.tls,
            )

        await self.streams[request.stream_id].handle(
            Request(
                stream_id=request.stream_id,
                headers=filter_pseudo_headers(request.headers),
                http_version="3",
                method=method,
                raw_path=raw_path,
                state=self.state,
            )
        )
        await self.context.mark_request()

    async def _create_server_push(
        self, stream_id: int, path: bytes, headers: list[tuple[bytes, bytes]]
    ) -> None:
        request_headers = [(b":method", b"GET"), (b":path", path)]
        request_headers.extend(headers)
        request_headers.extend(self.config.response_headers("h3"))
        try:
            push_stream_id = self.connection.send_push_promise(
                stream_id=stream_id, headers=request_headers
            )
        except NoAvailablePushIDError:
            # Client does not accept push promises or we are trying to
            # push on a push promises request.
            pass
        else:
            event = HeadersReceived(
                stream_id=push_stream_id, stream_ended=True, headers=request_headers
            )
            await self._create_stream(event)
            await self.streams[event.stream_id].handle(EndBody(stream_id=event.stream_id))



"""ASGI HTTP stream handler that maps between protocol events and ASGI messages."""

from __future__ import annotations

from enum import Enum, auto
from time import time
from typing import TYPE_CHECKING
from urllib.parse import unquote

from anycorn.typing import (
    AppWrapper,
    ASGISendEvent,
    Extensions,
    HTTPResponseStartEvent,
    HTTPScope,
    ResponseSummary,
    TaskGroup,
    TLSExtension,
    WorkerContext,
)
from anycorn.utils import (
    UnexpectedMessageError,
    build_and_validate_headers,
    suppress_body,
    valid_server_name,
)

from .events import (
    Body,
    EndBody,
    Event,
    InformationalResponse,
    Request,
    Response,
    StreamClosed,
    Trailers,
)

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable

    from anycorn.config import Config

TRAILERS_VERSIONS = {"2", "3"}
PUSH_VERSIONS = {"2", "3"}
EARLY_HINTS_VERSIONS = {"2", "3"}


class ASGIHTTPState(Enum):
    """State machine for ASGI HTTP response lifecycle.

    The ASGI Spec is clear that a response should not start till the
    framework has sent at least one body message hence why this
    state tracking is required.
    """

    REQUEST = auto()
    RESPONSE = auto()
    TRAILERS = auto()
    CLOSED = auto()


class HTTPStream:
    """Handles a single HTTP request/response stream."""

    def __init__(  # noqa: PLR0913
        self,
        app: AppWrapper,
        config: Config,
        context: WorkerContext,
        task_group: TaskGroup,
        client: tuple[str, int] | None,
        server: tuple[str, int] | None,
        send: Callable[[Event], Awaitable[None]],
        stream_id: int,
        tls: TLSExtension | None,
    ) -> None:
        """Initialize the HTTP stream handler."""
        self.app = app
        self.client = client
        self.closed = False
        self.config = config
        self.context = context
        self.response: HTTPResponseStartEvent
        self.scope: HTTPScope
        self.send = send
        self.server = server
        self.start_time: float
        self.state = ASGIHTTPState.REQUEST
        self.stream_id = stream_id
        self.task_group = task_group
        self.tls = tls
        self.scheme = "https" if self.tls is not None else "http"

    @property
    def idle(self) -> bool:
        """Return True when the stream is not actively processing a request."""
        return False

    async def handle(self, event: Event) -> None:  # noqa: C901, PLR0912
        """Handle an incoming stream event."""
        if self.closed:
            return
        if isinstance(event, Request):
            self.start_time = time()
            path, _, query_string = event.raw_path.partition(b"?")

            extensions = Extensions()
            if self.tls is not None:
                extensions["tls"] = self.tls

            if event.http_version in TRAILERS_VERSIONS:
                extensions["http.response.trailers"] = {}

            if event.http_version in PUSH_VERSIONS:
                extensions["http.response.push"] = {}

            if event.http_version in EARLY_HINTS_VERSIONS:
                extensions["http.response.early_hint"] = {}

            self.scope = HTTPScope(
                type="http",
                http_version=event.http_version,
                asgi={"spec_version": "2.1", "version": "3.0"},
                method=event.method,
                scheme=self.scheme,
                path=unquote(path.decode("ascii")),
                raw_path=path,
                query_string=query_string,
                root_path=self.config.root_path,
                headers=event.headers,
                client=self.client,
                server=self.server,
                state=event.state,
                extensions=extensions,
            )

            if valid_server_name(self.config, event):
                self.app_put = await self.task_group.spawn_app(
                    self.app, self.config, self.scope, self.app_send
                )
            else:
                await self._send_error_response(404)
                self.closed = True

        elif isinstance(event, Body):
            await self.app_put(
                {"type": "http.request", "body": bytes(event.data), "more_body": True}
            )
        elif isinstance(event, EndBody):
            await self.app_put({"type": "http.request", "body": b"", "more_body": False})
        elif isinstance(event, StreamClosed):
            self.closed = True
            if self.state != ASGIHTTPState.CLOSED:
                await self.config.log.access(self.scope, None, time() - self.start_time)
            if self.app_put is not None:
                await self.app_put({"type": "http.disconnect"})

    async def app_send(self, message: ASGISendEvent | None) -> None:  # noqa: C901, PLR0912
        """Handle a message sent by the ASGI application."""
        if message is None:  # ASGI App has finished sending messages
            if not self.closed:
                # Cleanup if required
                if self.state == ASGIHTTPState.REQUEST:
                    await self._send_error_response(500)
                await self.send(StreamClosed(stream_id=self.stream_id))
        elif message["type"] == "http.response.start" and self.state == ASGIHTTPState.REQUEST:
            self.response = message
            headers = build_and_validate_headers(self.response.get("headers", []))
            await self.send(
                Response(
                    stream_id=self.stream_id,
                    headers=headers,
                    status_code=int(self.response["status"]),
                )
            )
            self.state = ASGIHTTPState.RESPONSE
        elif (
            message["type"] == "http.response.push" and self.scope["http_version"] in PUSH_VERSIONS
        ):
            if not isinstance(message["path"], str):
                msg = f"{message['path']} should be a str"
                raise TypeError(msg)
            headers = [(b":scheme", self.scope["scheme"].encode())]
            for name, value in self.scope["headers"]:
                if name == b"host":
                    headers.append((b":authority", value))
            headers.extend(build_and_validate_headers(message["headers"]))
            await self.send(
                Request(
                    stream_id=self.stream_id,
                    headers=headers,
                    http_version=self.scope["http_version"],
                    method="GET",
                    raw_path=message["path"].encode(),
                    state=self.scope["state"],
                )
            )
        elif (
            message["type"] == "http.response.early_hint"
            and self.scope["http_version"] in EARLY_HINTS_VERSIONS
            and self.state == ASGIHTTPState.REQUEST
        ):
            headers = [(b"link", bytes(link).strip()) for link in message["links"]]
            await self.send(
                InformationalResponse(
                    stream_id=self.stream_id,
                    headers=headers,
                    status_code=103,
                )
            )
        elif message["type"] == "http.response.body" and self.state == ASGIHTTPState.RESPONSE:
            if (
                not suppress_body(self.scope["method"], int(self.response["status"]))
                and message.get("body", b"") != b""
            ):
                await self.send(
                    Body(stream_id=self.stream_id, data=bytes(message.get("body", b"")))
                )

            if not message.get("more_body", False):
                if self.response.get("trailers", False):
                    self.state = ASGIHTTPState.TRAILERS
                else:
                    await self._send_closed()
        elif (
            message["type"] == "http.response.trailers"
            and self.scope["http_version"] in TRAILERS_VERSIONS
            and self.state == ASGIHTTPState.REQUEST
        ):
            for name, value in self.scope["headers"]:
                if name == b"te" and value == b"trailers":
                    headers = build_and_validate_headers(message["headers"])
                    self.response = HTTPResponseStartEvent(
                        type="http.response.start",
                        status=200,
                        headers=headers,
                    )
                    await self.send(
                        Response(
                            stream_id=self.stream_id,
                            headers=headers,
                            status_code=200,
                        )
                    )
                    self.state = ASGIHTTPState.TRAILERS
                    break

            if not message.get("more_trailers", False):
                await self._send_closed()
        elif (
            message["type"] == "http.response.trailers"
            and self.scope["http_version"] in TRAILERS_VERSIONS
            and self.state == ASGIHTTPState.TRAILERS
        ):
            for name, value in self.scope["headers"]:
                if name == b"te" and value == b"trailers":
                    headers = build_and_validate_headers(message["headers"])
                    await self.send(Trailers(stream_id=self.stream_id, headers=headers))
                    break

            if not message.get("more_trailers", False):
                await self._send_closed()
        else:
            raise UnexpectedMessageError(self.state, message["type"])

    async def _send_closed(self) -> None:
        await self.send(EndBody(stream_id=self.stream_id))
        self.state = ASGIHTTPState.CLOSED
        await self.config.log.access(self.scope, self.response, time() - self.start_time)
        await self.send(StreamClosed(stream_id=self.stream_id))

    async def _send_error_response(self, status_code: int) -> None:
        await self.send(
            Response(
                stream_id=self.stream_id,
                headers=[(b"content-length", b"0"), (b"connection", b"close")],
                status_code=status_code,
            )
        )
        await self.send(EndBody(stream_id=self.stream_id))
        self.state = ASGIHTTPState.CLOSED
        await self.config.log.access(
            self.scope, ResponseSummary(status=status_code, headers=[]), time() - self.start_time
        )



"""Command-line entry point for Anycorn."""

from __future__ import annotations

import ssl
import sys

import rich_click as click

from .config import Config
from .run import run


def _load_config(config_path: str | None) -> Config:
    if config_path is None:
        return Config()
    if config_path.startswith("python:"):
        return Config.from_object(config_path[len("python:") :])
    if config_path.startswith("file:"):
        return Config.from_pyfile(config_path[len("file:") :])
    return Config.from_toml(config_path)


@click.command(
    help="Start the server and dispatch to the APPLICATION as path.to.module:instance.path."
)
@click.argument(
    "application",
)
@click.option(
    "--access-logfile",
    help="The target location for the access log, use `-` for stdout",
)
@click.option(
    "--access-logformat",
    help="The log format for the access log, see help docs",
)
@click.option("--backlog", type=int, help="The maximum number of pending connections")
@click.option(
    "-b",
    "--bind",
    "binds",
    help="The TCP host/address to bind to. Should be either host:port, host, "
    "unix:path or fd://num, e.g. 127.0.0.1:5000, 127.0.0.1, "
    "unix:/tmp/socket or fd://33 respectively.",
    default=[],
    multiple=True,
)
@click.option(
    "--ca-certs",
    help="Path to the SSL CA certificate file",
)
@click.option(
    "--certfile",
    help="Path to the SSL certificate file",
)
@click.option(
    "--cert-reqs",
    type=int,
    help="See verify mode argument",
)
@click.option(
    "--ciphers",
    help="Ciphers to use for the SSL setup",
)
@click.option(
    "-c",
    "--config",
    help="Location of a TOML config file, or when prefixed with `file:` a Python file, "
    "or when prefixed with `python:` a Python module.",
)
@click.option(
    "--debug",
    help="Enable debug mode, i.e. extra logging and checks",
    is_flag=True,
)
@click.option(
    "--error-logfile",
    "--log-file",
    "error_logfile",
    help="The target location for the error log, use `-` for stderr",
)
@click.option(
    "--graceful-timeout",
    help="Time to wait after SIGTERM or Ctrl-C for any remaining requests (tasks) to complete.",
    type=int,
)
@click.option(
    "--read-timeout",
    help="Seconds to wait before timing out reads on TCP sockets",
    type=int,
)
@click.option(
    "--max-requests",
    help="Maximum number of requests a worker will process before restarting",
    type=int,
)
@click.option(
    "--max-requests-jitter",
    help="This jitter causes the max-requests per worker to be "
    "randomized by randint(0, max_requests_jitter)",
    type=int,
)
@click.option(
    "-g",
    "--group",
    help="Group to own any unix sockets.",
    type=int,
)
@click.option(
    "-k",
    "--worker-class",
    help="The type of worker to use. Options include asyncio and trio.",
    type=click.Choice(("asyncio", "trio")),
)
@click.option(
    "--keep-alive",
    help="Seconds to keep inactive connections alive for",
    type=int,
)
@click.option(
    "--keyfile",
    help="Path to the SSL key file",
)
@click.option(
    "--keyfile-password",
    help="Password to decrypt the SSL key file",
)
@click.option(
    "--insecure-bind",
    "insecure_binds",
    help="The TCP host/address to bind to. SSL options will not apply to these binds. "
    "See *bind* for formatting options. Care must be taken! See HTTP -> HTTPS redirection docs.",
    default=[],
    multiple=True,
)
@click.option(
    "--log-config",
    help="A Python logging configuration file. This can be prefixed with "
    "'json:' or 'toml:' to load the configuration from a file in "
    " that format. Default is the logging ini format.",
)
@click.option(
    "--log-level",
    help="The (error) log level, defaults to info",
)
@click.option(
    "-p",
    "--pid",
    help="Location to write the PID (Program ID) to.",
)
@click.option(
    "--reload",
    help="Enable automatic reloads on code changes",
    is_flag=True,
)
@click.option(
    "--root-path",
    help="The setting for the ASGI root_path variable",
)
@click.option(
    "--server-name",
    "server_names",
    help="The hostnames that can be served, requests to different hosts "
    "will be responded to with 404s.",
    default=[],
    multiple=True,
)
@click.option(
    "--statsd-host",
    help="The host:port of the statsd server",
)
@click.option(
    "--statsd-prefix",
    help="Prefix for all statsd messages",
    default="",
)
@click.option(
    "-m",
    "--umask",
    help="The permissions bit mask to use on any unix sockets.",
    type=int,
)
@click.option(
    "-u",
    "--user",
    help="User to own any unix sockets.",
    type=int,
)
@click.option(
    "--verify-mode",
    help="SSL verify mode for peer's certificate, see ssl.VerifyMode enum for possible values.",
    type=click.Choice(("CERT_NONE", "CERT_OPTIONAL", "CERT_REQUIRED")),
)
@click.option(
    "--websocket-ping-interval",
    help="If set this is the time in seconds between pings sent to the client. "
    "This can be used to keep the websocket connection alive.",
    type=int,
)
@click.option(
    "--websocket-permessage-deflate",
    help='Enable WebSocket "permessage-deflate" compression.',
    type=bool,
    default=True,
    show_default=True,
)
@click.option(
    "-w",
    "--workers",
    help="The number of workers to spawn and use",
    type=int,
)
def main(  # noqa: C901 PLR0913 PLR0912 PLR0915
    application: str,
    access_logfile: str | None,
    access_logformat: str | None,
    backlog: int | None,
    binds: list[str],
    ca_certs: str | None,
    certfile: str | None,
    cert_reqs: int | None,
    ciphers: str | None,
    config: str | None,
    debug: bool,  # noqa: FBT001
    error_logfile: str | None,
    graceful_timeout: int | None,
    read_timeout: int | None,
    max_requests: int | None,
    max_requests_jitter: int | None,
    group: int | None,
    worker_class: str | None,
    keep_alive: int | None,
    keyfile: str | None,
    keyfile_password: str | None,
    insecure_binds: list[str],
    log_config: str | None,
    log_level: str | None,
    pid: str | None,
    reload: bool,  # noqa: FBT001
    root_path: str | None,
    server_names: list[str],
    statsd_host: str | None,
    statsd_prefix: str,
    umask: int | None,
    user: int | None,
    verify_mode: str | None,
    websocket_ping_interval: int | None,
    websocket_permessage_deflate: bool,  # noqa: FBT001
    workers: int | None,
) -> int:
    """Configure and start the Anycorn server with the given options."""
    cfg = _load_config(config)
    cfg.application_path = application

    if log_level is not None:
        cfg.loglevel = log_level
    if access_logformat is not None:
        cfg.access_log_format = access_logformat
    if access_logfile is not None:
        cfg.accesslog = access_logfile
    if backlog is not None:
        cfg.backlog = backlog
    if ca_certs is not None:
        cfg.ca_certs = ca_certs
    if certfile is not None:
        cfg.certfile = certfile
    if cert_reqs is not None:
        cfg.cert_reqs = cert_reqs
    if ciphers is not None:
        cfg.ciphers = ciphers
    cfg.debug = debug
    if error_logfile is not None:
        cfg.errorlog = error_logfile
    if graceful_timeout is not None:
        cfg.graceful_timeout = graceful_timeout
    if read_timeout is not None:
        cfg.read_timeout = read_timeout
    if group is not None:
        cfg.group = group
    if keep_alive is not None:
        cfg.keep_alive_timeout = keep_alive
    if keyfile is not None:
        cfg.keyfile = keyfile
    if keyfile_password is not None:
        cfg.keyfile_password = keyfile_password
    if log_config is not None:
        cfg.logconfig = log_config
    if max_requests is not None:
        cfg.max_requests = max_requests
    if max_requests_jitter is not None:
        cfg.max_requests_jitter = max_requests_jitter
    if pid is not None:
        cfg.pid_path = pid
    if root_path is not None:
        cfg.root_path = root_path
    cfg.use_reloader = reload
    if statsd_host is not None:
        cfg.statsd_host = statsd_host
    if statsd_prefix is not None:
        cfg.statsd_prefix = statsd_prefix
    if umask is not None:
        cfg.umask = umask
    if user is not None:
        cfg.user = user
    if worker_class is not None:
        cfg.worker_class = worker_class
    if verify_mode is not None:
        cfg.verify_mode = ssl.VerifyMode[verify_mode]
    if websocket_ping_interval is not None:
        cfg.websocket_ping_interval = websocket_ping_interval
    cfg.websocket_permessage_deflate = websocket_permessage_deflate
    if workers is not None:
        cfg.workers = workers

    if len(binds) > 0:
        cfg.bind = binds
    if len(insecure_binds) > 0:
        cfg.insecure_bind = insecure_binds
    if len(server_names) > 0:
        cfg.server_names = server_names

    return run(cfg)


if __name__ == "__main__":
    sys.exit(main())



"""Internal stream-level events used between protocol layers and stream handlers."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from anycorn.typing import ConnectionState

_MIN_INFORMATIONAL_STATUS = 100
_MAX_INFORMATIONAL_STATUS = 200


@dataclass(frozen=True)
class Event:
    """Base class for all stream events."""

    stream_id: int


@dataclass(frozen=True)
class Request(Event):
    """Represents an incoming HTTP request."""

    headers: list[tuple[bytes, bytes]]
    http_version: str
    method: str
    raw_path: bytes
    state: ConnectionState


@dataclass(frozen=True)
class Body(Event):
    """Represents a chunk of request body data."""

    data: bytes


@dataclass(frozen=True)
class EndBody(Event):
    """Signals the end of the request body."""


@dataclass(frozen=True)
class Trailers(Event):
    """Represents HTTP trailers sent after the body."""

    headers: list[tuple[bytes, bytes]]


@dataclass(frozen=True)
class Data(Event):
    """Represents raw data (e.g. WebSocket frames)."""

    data: bytes


@dataclass(frozen=True)
class EndData(Event):
    """Signals the end of a data stream (e.g. WebSocket close)."""


@dataclass(frozen=True)
class Response(Event):
    """Represents an HTTP response to be sent."""

    headers: list[tuple[bytes, bytes]]
    status_code: int


@dataclass(frozen=True)
class InformationalResponse(Event):
    """Represents a 1XX informational HTTP response."""

    headers: list[tuple[bytes, bytes]]
    status_code: int

    def __post_init__(self) -> None:
        """Validate that the status code is in the 1XX range."""
        if (
            self.status_code >= _MAX_INFORMATIONAL_STATUS
            or self.status_code < _MIN_INFORMATIONAL_STATUS
        ):
            msg = f"Status code must be 1XX not {self.status_code}"
            raise ValueError(msg)


@dataclass(frozen=True)
class StreamClosed(Event):
    """Signals that the stream has been closed."""



"""Utility functions and exception types for the Anycorn server."""

from __future__ import annotations

import contextlib
import functools
import inspect
import os
import pathlib
import re
import socket
import ssl
import sys
from importlib import import_module
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    cast,
)

from anyio import TypedAttributeLookupError
from anyio.streams.tls import TLSAttribute

from .app_wrappers import ASGIWrapper, WSGIWrapper

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable, Iterable
    from enum import Enum
    from multiprocessing.synchronize import Event as EventType

    from anyio.abc import SocketStream

    from .config import Config
    from .protocol.events import Request
    from .typing import AppWrapper, ASGIFramework, Framework, TLSExtension, WSGIFramework


class ShutdownError(Exception):
    """Raised to trigger a graceful server shutdown."""


class NoAppError(Exception):
    """Raised when the application cannot be loaded."""


class LifespanTimeoutError(Exception):
    """Raised when the ASGI lifespan startup or shutdown times out."""

    def __init__(self, stage: str) -> None:
        super().__init__(
            f"Timeout whilst awaiting {stage}. Your application may not support the ASGI Lifespan "
            f"protocol correctly, alternatively the {stage}_timeout configuration is incorrect."
        )


class LifespanFailureError(Exception):
    """Raised when the ASGI lifespan reports a failure."""

    def __init__(self, stage: str, message: str) -> None:
        super().__init__(f"Lifespan failure in {stage}. '{message}'")


class UnexpectedMessageError(Exception):
    """Raised when an unexpected ASGI message type is received for the current state."""

    def __init__(self, state: Enum, message_type: str) -> None:
        super().__init__(f"Unexpected message type, {message_type} given the state {state}")


class FrameTooLargeError(Exception):
    """Raised when a WebSocket frame exceeds the maximum allowed size."""


_CERT_PATTERN = re.compile(r"-----BEGIN CERTIFICATE-----\s.*?-----END CERTIFICATE-----", re.DOTALL)

_TLS_VERSION_PREFIX = "TLSv"

_TLS_VERSION_MAP = {
    "SSLv3": 0x0300,
    "TLSv1": 0x0301,
    "TLSv1.1": 0x0302,
    "TLSv1.2": 0x0303,
    "TLSv1.3": 0x0304,
}

_SERVER_CERT_CACHE: dict[str, str | None] = {}

RFC4514_ATTRIBUTE_NAMES = {
    "commonName": "CN",
    "countryName": "C",
    "localityName": "L",
    "stateOrProvinceName": "ST",
    "organizationName": "O",
    "organizationalUnitName": "OU",
    "emailAddress": "emailAddress",
    "serialNumber": "serialNumber",
    "streetAddress": "street",
    "postalCode": "postalCode",
    "domainComponent": "DC",
}

TLS_CIPHER_NAME_TO_CODE: dict[str, int] = {
    # TLS 1.3 identifiers (RFC 8446, Appendix B.4)
    "TLS_AES_128_GCM_SHA256": 0x1301,
    "TLS_AES_256_GCM_SHA384": 0x1302,
    "TLS_CHACHA20_POLY1305_SHA256": 0x1303,
    "TLS_AES_128_CCM_SHA256": 0x1304,
    "TLS_AES_128_CCM_8_SHA256": 0x1305,
    # Common TLS 1.2 suites (IANA TLS Cipher Suite Registry)
    "ECDHE-RSA-AES128-GCM-SHA256": 0xC02F,
    "ECDHE-RSA-AES256-GCM-SHA384": 0xC030,
    "ECDHE-ECDSA-AES128-GCM-SHA256": 0xC02B,
    "ECDHE-ECDSA-AES256-GCM-SHA384": 0xC02C,
    "ECDHE-RSA-CHACHA20-POLY1305": 0xCCA8,
    "ECDHE-ECDSA-CHACHA20-POLY1305": 0xCCA9,
    "DHE-RSA-AES128-GCM-SHA256": 0x009E,
    "DHE-RSA-AES256-GCM-SHA384": 0x009F,
    "AES128-GCM-SHA256": 0x009C,
    "AES256-GCM-SHA384": 0x009D,
    "ECDHE-RSA-AES128-SHA256": 0xC027,
    "ECDHE-RSA-AES256-SHA384": 0xC028,
    "ECDHE-ECDSA-AES128-SHA256": 0xC023,
    "ECDHE-ECDSA-AES256-SHA384": 0xC024,
}


def default_tls_extension() -> TLSExtension:
    """Return a TLSExtension dict with all fields set to their default values."""
    return {
        "server_cert": None,
        "client_cert_chain": (),
        "client_cert_name": None,
        "client_cert_error": None,
        "tls_version": None,
        "cipher_suite": None,
    }


def tls_version_to_int(value: str | int | None) -> int | None:
    """Convert a TLS version string or integer to its IANA numeric code."""
    if isinstance(value, int):
        return value
    if isinstance(value, str):
        mapping = _TLS_VERSION_MAP.get(value)
        if mapping is not None:
            return mapping
        if value.startswith(_TLS_VERSION_PREFIX):
            version = value[len(_TLS_VERSION_PREFIX) :]
            if version == "1":
                return _TLS_VERSION_MAP["TLSv1"]
            try:
                major, _, minor = version.partition(".")
                if major == "1" and minor.isdigit():
                    return 0x0300 + int(minor) + 1
            except ValueError:
                return None
    return None


@functools.lru_cache(maxsize=16)
def _cached_server_cert(path: str) -> str | None:
    try:
        text = pathlib.Path(path).read_text(encoding="utf-8")
    except (OSError, UnicodeError):
        return None
    match = _CERT_PATTERN.search(text)
    if match is None:
        return None
    return match.group(0)


def get_server_certificate_pem(config: Config) -> str | None:
    """Return the PEM-encoded server certificate from config, or None."""
    if config.certfile is None:
        return None
    path = str(pathlib.Path(config.certfile).resolve())
    cached = _SERVER_CERT_CACHE.get(path)
    if cached is not None:
        return cached
    cert = _cached_server_cert(path)
    _SERVER_CERT_CACHE[path] = cert
    return cert


def _extract_client_chain(ssl_object: ssl.SSLObject) -> tuple[str, ...]:
    chain_der: tuple[bytes, ...] | tuple[()] = ()
    for method_name in ("get_verified_chain", "get_unverified_chain"):
        method = getattr(ssl_object, method_name, None)
        if callable(method):
            try:
                data = method()
            except ssl.SSLError:
                continue
            else:
                if data:
                    chain_der = tuple(data)
                    break
    if not chain_der:
        try:
            peer_cert = ssl_object.getpeercert(binary_form=True)
        except ssl.SSLError:
            peer_cert = None
        if peer_cert:
            chain_der = (peer_cert,)

    return tuple(ssl.DER_cert_to_PEM_cert(cert) for cert in chain_der if cert)


def _escape_rfc4514_value(value: str) -> str:
    escaped = []
    for index, char in enumerate(value):
        if (
            char in {",", "+", '"', "\\", "<", ">", ";"}
            or (index == 0 and char in {"#", " "})
            or (index == len(value) - 1 and char == " ")
        ):
            escaped.append("\\" + char)
        else:
            escaped.append(char)
    return "".join(escaped)


def _subject_to_rfc4514(subject: Iterable[Iterable[tuple[str, str]]]) -> str | None:
    rdns: list[str] = []
    try:
        for rdn in reversed(tuple(subject)):
            attrs = []
            for key, value in rdn:
                name = RFC4514_ATTRIBUTE_NAMES.get(key, key)
                attrs.append(f"{name}={_escape_rfc4514_value(str(value))}")
            rdns.append("+".join(attrs))
    except Exception:  # noqa: BLE001
        return None
    return ",".join(rdns) if rdns else None


def build_tls_extension(config: Config, stream: SocketStream) -> TLSExtension:  # noqa: C901, PLR0912
    """Extract TLS parameters for the ASGI connection scope.

    Harvest information from AnyIO's TLS attributes and Python's ``ssl`` library:

    * TLS version / cipher IDs follow RFC 8446 (TLS 1.3) and the IANA registry.
    * Client certificate chains are returned as PEM per ASGI spec.
    * Distinguished names are rendered using RFC 4514 formatting.

    References:
      - ASGI TLS Extension: https://asgi.readthedocs.io/en/latest/specs/tls.html
      - RFC 8446: https://www.rfc-editor.org/rfc/rfc8446
      - IANA TLS Cipher Suite Registry: https://www.iana.org/assignments/tls-parameters/tls-parameters.xhtml#tls-parameters-4
      - RFC 4514 (LDAP DN string format): https://www.rfc-editor.org/rfc/rfc4514

    """
    extension = default_tls_extension()

    server_cert = get_server_certificate_pem(config)
    if server_cert is not None:
        extension["server_cert"] = server_cert

    try:
        tls_version_value = stream.extra(TLSAttribute.tls_version)  # noqa: S610
    except TypedAttributeLookupError:
        tls_version_value = None
    extension["tls_version"] = tls_version_to_int(tls_version_value)

    try:
        ssl_object = stream.extra(TLSAttribute.ssl_object)  # noqa: S610
    except (TypedAttributeLookupError, AttributeError):
        ssl_object = None

    if ssl_object is not None:
        client_chain = _extract_client_chain(ssl_object)
        extension["client_cert_chain"] = client_chain
        extension["cipher_suite"] = None
        extension["client_cert_error"] = None

        try:
            cipher = ssl_object.cipher()
        except (AttributeError, ssl.SSLError):
            cipher = None
        if cipher:
            extension["cipher_suite"] = TLS_CIPHER_NAME_TO_CODE.get(cipher[0])

        try:
            context = ssl_object.context
        except AttributeError:
            context = None
        else:
            if (
                not client_chain
                and extension["client_cert_name"] is None
                and getattr(context, "verify_mode", ssl.CERT_NONE) == ssl.CERT_REQUIRED
            ):
                extension["client_cert_error"] = "missing-client-certificate"
    else:
        extension["client_cert_chain"] = ()

    if extension["client_cert_name"] is None:
        try:
            peer_certificate = stream.extra(TLSAttribute.peer_certificate)  # noqa: S610
        except TypedAttributeLookupError:
            peer_certificate = None
        if peer_certificate and "subject" in peer_certificate:
            extension["client_cert_name"] = _subject_to_rfc4514(
                cast("Iterable[Iterable[tuple[str, str]]]", peer_certificate["subject"])
            )

    if extension["client_cert_name"] is None and isinstance(ssl_object, ssl.SSLObject):
        try:
            cert_dict = ssl_object.getpeercert()
        except (ssl.SSLError, ValueError):
            cert_dict = None
        if cert_dict and "subject" in cert_dict:
            extension["client_cert_name"] = _subject_to_rfc4514(
                cast("Iterable[Iterable[tuple[str, str]]]", cert_dict["subject"])
            )

    if extension["client_cert_chain"] and not extension["client_cert_name"]:
        extension["client_cert_name"] = None

    return extension


def suppress_body(method: str, status_code: int) -> bool:
    """Return True if the response body should be suppressed for this method/status."""
    return method == "HEAD" or 100 <= status_code < 200 or status_code in {204, 304}  # noqa: PLR2004


def build_and_validate_headers(
    headers: Iterable[tuple[bytes, bytes]],
) -> list[tuple[bytes, bytes]]:
    """Validate and normalise response headers, raising ValueError for pseudo-headers."""
    # Validates that the header name and value are bytes
    validated_headers: list[tuple[bytes, bytes]] = []
    for name, value in headers:
        if name[0] == b":"[0]:
            msg = "Pseudo headers are not valid"
            raise ValueError(msg)
        validated_headers.append((bytes(name).strip(), bytes(value).strip()))
    return validated_headers


def filter_pseudo_headers(headers: list[tuple[bytes, bytes]]) -> list[tuple[bytes, bytes]]:
    """Remove HTTP/2 pseudo-headers and promote :authority to the host header."""
    filtered_headers: list[tuple[bytes, bytes]] = [(b"host", b"")]  # Placeholder
    authority = None
    host = b""
    for name, value in headers:
        if name == b":authority":  # h2 & h3 libraries validate this is present
            authority = value
        elif name == b"host":
            host = value
        elif name[0] != b":"[0]:
            filtered_headers.append((name, value))
    filtered_headers[0] = (b"host", authority if authority is not None else host)
    return filtered_headers


def load_application(path: str, wsgi_max_body_size: int) -> AppWrapper:
    """Load and wrap an application from a module path string."""
    mode: Literal["asgi", "wsgi"] | None = None
    if ":" not in path:
        module_name, app_name = path, "app"
    elif path.count(":") == 2:  # noqa: PLR2004
        mode, module_name, app_name = path.split(":", 2)  # type: ignore[assignment]
        if mode not in {"asgi", "wsgi"}:
            msg = "Invalid mode, must be 'asgi', or 'wsgi'"
            raise ValueError(msg)
    else:
        module_name, app_name = path.split(":", 1)

    module_path = pathlib.Path(module_name).resolve()
    sys.path.insert(0, str(module_path.parent))
    import_name = module_path.with_suffix("").name if module_path.is_file() else module_path.name
    try:
        module = import_module(import_name)
    except ModuleNotFoundError as error:
        if error.name == import_name:
            msg = f"Cannot load application from '{path}', module not found."
            raise NoAppError(msg) from error
        raise
    try:
        app = eval(app_name, vars(module))  # noqa: S307
    except NameError as error:
        msg = f"Cannot load application from '{path}', application not found."
        raise NoAppError(msg) from error
    else:
        return wrap_app(app, wsgi_max_body_size, mode)


def wrap_app(
    app: Framework, wsgi_max_body_size: int, mode: Literal["asgi", "wsgi"] | None
) -> AppWrapper:
    """Wrap *app* in the appropriate ASGI or WSGI wrapper."""
    if mode is None:
        mode = "asgi" if is_asgi(app) else "wsgi"
    if mode == "asgi":
        return ASGIWrapper(cast("ASGIFramework", app))
    return WSGIWrapper(cast("WSGIFramework", app), wsgi_max_body_size)


def files_to_watch() -> dict[pathlib.Path, float]:
    """Return a mapping of loaded module file paths to their current mtime."""
    last_updates: dict[pathlib.Path, float] = {}
    for module in list(sys.modules.values()):
        filename = getattr(module, "__file__", None)
        if filename is None:
            continue
        path = pathlib.Path(filename)
        with contextlib.suppress(FileNotFoundError, NotADirectoryError):
            last_updates[pathlib.Path(filename)] = path.stat().st_mtime
    return last_updates


def check_for_updates(files: dict[pathlib.Path, float]) -> bool:
    """Return True if any watched file has been modified since last checked."""
    for path, last_mtime in files.items():
        try:
            mtime = path.stat().st_mtime
        except FileNotFoundError:  # noqa: PERF203
            return True
        else:
            if mtime > last_mtime:
                return True
            files[path] = mtime
    return False


async def raise_shutdown(shutdown_event: Callable[..., Awaitable]) -> None:
    """Await *shutdown_event* then raise ShutdownError to trigger graceful shutdown."""
    await shutdown_event()
    raise ShutdownError


async def check_multiprocess_shutdown_event(
    shutdown_event: EventType, sleep: Callable[[float], Awaitable[Any]]
) -> None:
    """Poll *shutdown_event* in a loop, returning when it becomes set."""
    while True:
        if shutdown_event.is_set():
            return
        await sleep(0.1)


def write_pid_file(pid_path: str) -> None:
    """Write the current process ID to *pid_path*."""
    with pathlib.Path(pid_path).open("w") as file_:
        file_.write(f"{os.getpid()}")


def parse_socket_addr(family: int, address: tuple) -> tuple[str, int] | None:
    """Parse a raw socket address into a (host, port) tuple, or None for unsupported families."""
    if family == socket.AF_INET:
        return address
    if family == socket.AF_INET6:
        return (address[0], address[1])
    return None


def repr_socket_addr(family: int, address: tuple) -> str:
    """Return a human-readable string representation of a socket address."""
    if family == socket.AF_INET:
        return f"{address[0]}:{address[1]}"
    if family == socket.AF_INET6:
        return f"[{address[0]}]:{address[1]}"
    if family == socket.AF_UNIX:
        return f"unix:{address}"
    return f"{address}"


def valid_server_name(config: Config, request: Request) -> bool:
    """Return True if the request's Host header matches a configured server name."""
    if len(config.server_names) == 0:
        return True

    host = ""
    for name, value in request.headers:
        if name.lower() == b"host":
            host = value.decode()
            break
    return host in config.server_names


def is_asgi(app: Any) -> bool:  # noqa: ANN401
    """Return True if *app* appears to be an ASGI application."""
    if inspect.iscoroutinefunction(app):
        return True
    if callable(app):
        return inspect.iscoroutinefunction(app.__call__)
    return False



"""TCP server implementation for handling incoming connections."""

from __future__ import annotations

import contextlib
from math import inf
from ssl import SSLError, SSLZeroReturnError
from typing import TYPE_CHECKING

import anyio
import anyio.abc
import anyio.streams.tls

from .events import Closed, Event, RawData, Updated
from .protocol import ProtocolWrapper
from .task_group import TaskGroup
from .typing import AppWrapper, ConnectionState, LifespanState
from .utils import build_tls_extension, parse_socket_addr
from .worker_context import AnyioSingleTask, WorkerContext

if TYPE_CHECKING:
    from collections.abc import Callable

    from .config import Config

MAX_RECV = 2**16


class TCPServer:
    """Handles a single TCP connection, managing protocol negotiation and I/O."""

    def __init__(
        self,
        app: AppWrapper,
        config: Config,
        context: WorkerContext,
        state: LifespanState,
        stream: anyio.abc.SocketStream,
    ) -> None:
        self.app = app
        self.config = config
        self.context = context
        self.protocol: ProtocolWrapper
        self.send_lock = anyio.Lock()
        self.idle_task = AnyioSingleTask()
        self.state = state
        self.stream = stream

        self._idle_handle: anyio.CancelScope | None = None

    async def run(self) -> None:
        """Run the server for this connection."""
        try:
            alpn_protocol = self.stream.extra(anyio.streams.tls.TLSAttribute.alpn_protocol)  # noqa: S610
            tls_extension = build_tls_extension(self.config, self.stream)
        except anyio.TypedAttributeLookupError:  # Not SSL
            alpn_protocol = "http/1.1"
            tls_extension = None

        try:
            socket = self.stream.extra(anyio.abc.SocketAttribute.raw_socket)  # noqa: S610
            client = parse_socket_addr(socket.family, socket.getpeername())
            server = parse_socket_addr(socket.family, socket.getsockname())

            async with TaskGroup() as task_group:
                self._task_group = task_group
                self.protocol = ProtocolWrapper(
                    self.app,
                    self.config,
                    self.context,
                    task_group,
                    ConnectionState(self.state.copy()),
                    client,
                    server,
                    self.protocol_send,
                    tls_extension,
                    alpn_protocol,
                )
                await self.protocol.initiate()
                await self.idle_task.restart(self._task_group, self._idle_timeout)
                await self._read_data()
        except OSError:
            pass
        finally:
            await self._close()

    async def protocol_send(self, event: Event) -> None:
        """Forward a protocol event to the underlying stream."""
        if isinstance(event, RawData):
            async with self.send_lock:
                try:
                    with anyio.CancelScope(shield=True):
                        await self.stream.send(event.data)
                except (anyio.ClosedResourceError, anyio.BrokenResourceError, TimeoutError):
                    await self.protocol.handle(Closed())
        elif isinstance(event, Closed):
            await self._close()
            await self.protocol.handle(Closed())
        elif isinstance(event, Updated):
            if event.idle:
                await self.idle_task.restart(self._task_group, self._idle_timeout)
            else:
                await self.idle_task.stop()

    async def _read_data(self) -> None:
        while True:
            try:
                with anyio.fail_after(self.config.read_timeout or inf):
                    data = await self.stream.receive(MAX_RECV)
            except (  # noqa: PERF203
                anyio.ClosedResourceError,
                anyio.BrokenResourceError,
                anyio.EndOfStream,
                TimeoutError,
                SSLZeroReturnError,
            ):
                break
            else:
                await self.protocol.handle(RawData(data))
                if data == b"":
                    break
        await self.protocol.handle(Closed())

    async def _close(self) -> None:
        with contextlib.suppress(
            OSError,
            anyio.BrokenResourceError,
            AttributeError,
            NotImplementedError,
            TypeError,
            anyio.BusyResourceError,
            anyio.ClosedResourceError,
        ):
            # They're already gone, nothing to do, or it is a SSL stream
            await self.stream.send_eof()
        with contextlib.suppress(
            SSLError, anyio.ClosedResourceError, anyio.BrokenResourceError, anyio.BusyResourceError
        ):
            await self.stream.aclose()

    async def _idle_timeout(self) -> None:
        with anyio.move_on_after(self.config.keep_alive_timeout):
            await self.context.terminated.wait()

        with anyio.CancelScope(shield=True):
            await self._initiate_server_close()

    async def _initiate_server_close(self) -> None:
        await self.protocol.handle(Closed())
        with contextlib.suppress(SSLError, anyio.BrokenResourceError, anyio.BusyResourceError):
            await self.stream.aclose()


def tcp_server_handler(
    app: AppWrapper,
    config: Config,
    context: WorkerContext,
    state: LifespanState,
) -> Callable:
    """Return a handler callable suitable for use with anyio's listener.serve()."""

    async def handler(stream: anyio.abc.SocketStream) -> None:
        tcp_server = TCPServer(app, config, context, state, stream)
        await tcp_server.run()

    return handler



"""Task group implementation wrapping anyio for concurrent ASGI request handling."""

from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Any

import anyio
import anyio.abc
import anyio.from_thread
import anyio.to_thread

from .typing import AppWrapper, ASGIReceiveCallable, ASGIReceiveEvent, ASGISendEvent, Scope

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable
    from types import TracebackType

    from .config import Config

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from exceptiongroup import BaseExceptionGroup
    from typing_extensions import Self


async def _handle(  # noqa: PLR0913
    app: AppWrapper,
    config: Config,
    scope: Scope,
    receive: ASGIReceiveCallable,
    send: Callable[[ASGISendEvent | None], Awaitable[None]],
    sync_spawn: Callable,
    call_soon: Callable,
) -> None:
    try:
        await app(scope, receive, send, sync_spawn, call_soon)
    except anyio.get_cancelled_exc_class():
        raise
    except BaseExceptionGroup as error:
        _, other_errors = error.split(anyio.get_cancelled_exc_class())
        if other_errors is not None:
            await config.log.exception("Error in ASGI Framework")
            await send(None)
        else:
            raise
    except Exception:  # noqa: BLE001
        await config.log.exception("Error in ASGI Framework")
    finally:
        await send(None)


class TaskGroup:
    """Manages concurrent ASGI app tasks using an anyio task group."""

    def __init__(self) -> None:
        self._task_group: anyio.abc.TaskGroup | None = None

    async def spawn_app(
        self,
        app: AppWrapper,
        config: Config,
        scope: Scope,
        send: Callable[[ASGISendEvent | None], Awaitable[None]],
    ) -> Callable[[ASGIReceiveEvent], Awaitable[None]]:
        """Spawn an ASGI app task and return a callable to send receive events to it."""
        app_send_channel, app_receive_channel = anyio.create_memory_object_stream[ASGIReceiveEvent](
            config.max_app_queue_size
        )
        assert self._task_group is not None
        self._task_group.start_soon(
            _handle,
            app,
            config,
            scope,
            app_receive_channel.receive,
            send,
            anyio.to_thread.run_sync,
            anyio.from_thread.run,
        )
        return app_send_channel.send

    def spawn(self, func: Callable, *args: Any) -> None:  # noqa: ANN401
        """Spawn an arbitrary coroutine function in the task group."""
        assert self._task_group is not None
        self._task_group.start_soon(func, *args)

    async def __aenter__(self) -> Self:
        tg = anyio.create_task_group()
        self._task_group = await tg.__aenter__()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        tb: TracebackType | None,
    ) -> None:
        assert self._task_group is not None
        await self._task_group.__aexit__(exc_type, exc_value, tb)
        self._task_group = None



"""Protocol wrapper that selects and manages the appropriate HTTP protocol handler."""

from __future__ import annotations

from typing import TYPE_CHECKING

from anycorn.events import Event, RawData

from .h2 import H2Protocol
from .h11 import H2CProtocolRequiredError, H2ProtocolAssumedError, H11Protocol

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable

    from anycorn.config import Config
    from anycorn.typing import AppWrapper, ConnectionState, TaskGroup, TLSExtension, WorkerContext


class ProtocolWrapper:
    """Wraps H11 and H2 protocols, upgrading as needed based on ALPN or upgrade headers."""

    def __init__(  # noqa: PLR0913
        self,
        app: AppWrapper,
        config: Config,
        context: WorkerContext,
        task_group: TaskGroup,
        state: ConnectionState,
        client: tuple[str, int] | None,
        server: tuple[str, int] | None,
        send: Callable[[Event], Awaitable[None]],
        tls: TLSExtension | None,
        alpn_protocol: str | None = None,
    ) -> None:
        """Initialize the protocol wrapper."""
        self.app = app
        self.config = config
        self.context = context
        self.task_group = task_group
        self.client = client
        self.server = server
        self.send = send
        self.state = state
        self.tls = tls
        self.protocol: H11Protocol | H2Protocol
        if alpn_protocol == "h2":
            self.protocol = H2Protocol(
                self.app,
                self.config,
                self.context,
                self.task_group,
                self.state,
                self.client,
                self.server,
                self.send,
                self.tls,
            )
        else:
            self.protocol = H11Protocol(
                self.app,
                self.config,
                self.context,
                self.task_group,
                self.state,
                self.client,
                self.server,
                self.send,
                self.tls,
            )

    async def initiate(self) -> None:
        """Initiate the underlying protocol."""
        return await self.protocol.initiate()

    async def handle(self, event: Event) -> None:
        """Handle an incoming event, upgrading the protocol if required."""
        try:
            return await self.protocol.handle(event)
        except H2ProtocolAssumedError as error:
            self.protocol = H2Protocol(
                self.app,
                self.config,
                self.context,
                self.task_group,
                self.state,
                self.client,
                self.server,
                self.send,
                self.tls,
            )
            await self.protocol.initiate()
            if error.data != b"":
                return await self.protocol.handle(RawData(data=error.data))
        except H2CProtocolRequiredError as error:
            self.protocol = H2Protocol(
                self.app,
                self.config,
                self.context,
                self.task_group,
                self.state,
                self.client,
                self.server,
                self.send,
                self.tls,
            )
            await self.protocol.initiate(error.headers, error.settings)
            if error.data != b"":
                return await self.protocol.handle(RawData(data=error.data))
